'use strict';

/**
 * This is a DOM based semantic templating engine inspired by Angular and Transparency
 */

/**
 * Cheerio Bugs
 * - attr to null doesn't clear the attribute
 * - self closing tags can mess up .replaceWith, it'll start blowing up it's siblings
 */

const cloneDeep = require('lodash.clonedeep')
const isPlainObject = require('lodash.isplainobject')
const merge = require('lodash.merge')

/**
 * Simple shim of the jQuery selector syntax
 * - it doesn't support need to support find or collection operations like jQuery, actually an anti pattern in this usecase
 * - it just extends the original react params object with added functionality
 * @param node
 * @returns $node
 */
const $ = (node) => {
  let $node = new Query(node)
  return $node
}

function Query(node) {
  Object.assign(this, node || {})

  this.$parent = null
  return this
}

Query.prototype.findBinding = function(key) {
  let bindKey = this.attr('data-bind') || this.attr('id')

  if (bindKey === key)
    return this
  else
    return this.getChildren().find($child => {
      if ($child.findBinding)
        return $child.findBinding(key)
    })
}

Query.prototype.findClass = function(className) {
  let klass = this.props.className || ''
  if (klass.split(/\s+/).indexOf(className) >= 0)
    return this
  else
    return this.getChildren().find($child => {
      if ($child.findClass)
        return $child.findClass(className)
    })
}

Query.prototype.clone = function() {
  let cloneNode = cloneDeep(this)

  return $(cloneNode)
}


Query.prototype.css = function(style) {
  if (style === undefined)
    return this.props.style

  else
    this.props.style = style

  return this
}

Query.prototype.attr = function(attr, setValue) {
  if (attr === undefined)
    return this.props
  else {
    if (setValue !== undefined) {
      this.props[attr] == setValue
    } else {
      return this.props[attr]
    }
  }
}

Query.prototype.text = function(text) {
  if (text === undefined)
    return (this.children || [])
      .filter(child => typeof child === 'text')
      .join('')
  else {
    this.children = [text]
    return this
  }
}

Query.prototype.parent = function() {
  return this.$parent || new Query
}

// NOTE: This can't be named children as it clashes with the VDOM structure
Query.prototype.getChildren = function() {
  let children = this.children || []

  if (!(children instanceof Array))
    children = [children]

  return children.map(child => {
    if (child && child.type) {
      let $child = new Query(child)
      $child.$parent = this
      return $child

    } else {
      return child
    }
  })
}

Query.prototype.siblings = function() {
  return this.parent().getChildren()
}

Query.prototype.html = function(newNodes) {
  //TODO flatten the newNodes
  this.children = newNodes
  return this
}

Query.prototype.replaceWith = function($node) {
  if ($node === undefined) return this
  //if ($node instanceof Array) {
  //  // NOTE: jquery's replace is just a remove and append
  //  let siblings = this.siblings()
  //  let index = siblings.indexOf(this)
  //
  //  if (index >= 0)
  //    Array.prototype.splice.apply(this.$parent.children, [0, $node.length].concat($node))
  //  //this.node = $node[0]
  //  //$node.forEach($node => {
  //  //  this.after($node)
  //  //})
  //
  //  return this
  //}

  if (!($node instanceof Query)) throw 'replaceWith must have a Query instance'

  // Don't merge in parent relationships
  delete $node.$parent
  Object.assign(this, $node)

  return this
}


/**
 * By default the directive is matched if the key is the same, however you can provide custom matcher
 * - the short form is to just pass in a named render function, the function name will be used to match the binding
 * - NOTE: the order in which the directives run matters
 */
const defaultDirectives = {

  _array: {
    recursive: true,
    match: function (value, key) {
      return value && value instanceof Array
    },
    render: function ($target, value, context) {
      // NOTE: this should only support data-bind since id should not be cloned

      // NOTE: a closure won't be created with argument variables, need to create another var
      const key = context.key,
        basePath = context.path

      const cacheKey = compact([basePath, key]).join('.')

      let $elements = value.map((data, i) => {
        let path = compact([basePath, key, i]).join('.')
        let $template = this.memoized(cacheKey, $target)

        return this._render($template, data, path)
      })

      return $elements
    }
  },

  _object: {
    recursive: true,
    match: function (value, key) {
      return isPlainObject(value)
    },
    render: function ($target, value, context) {
      // NOTE: doesn't need to be memoized, it's done on the top level already
      return this._render($target, value, compact([context.path, context.key]).join('.'))
    }
  },

  _text: {
    // NOTE if the match conditions don't overlap, then order doesn't really matter
    // refactor the directives as an object so that it's easy to overwrite certain properties
    // This is just the default directive
    match: function (value, key) {
      return typeof value != 'object' || value == null
    },
    render: function text($target, value, context) {

      return $target.text(value)
    }
  }
}


/**
 * Default meta directives
 * - meta directives runs before normal directives and does not require a $target to bind to
 */
const metaDirectives = {
  _template: function($node, value) {
    let templateKey = value
    let $root = this.$node
    let $template = $root.findBinding(templateKey)
    if (!$template)
      return console.error('$template not found', templateKey)
    let $content = $template.getChildren()
    //$node.props['className'] = templateKey
    $node.html($content)
  },

  _html: function ($node, value) {
    return $node.html(value)
  },

  _id: function attributes($target, value) {
    $target.props['id'] = value
    return $target
  },

  // TODO implement & allow all react attributes
  _className: function attributes($target, value, context) {
    $target.props['className'] = $target.props.className ? $target.props.className + ' ' + value : value
    return $target
  }
}

/**
 * @param html - an html snippet
 * @param directives
 * @param onRender
 * @constructor
 */
function Bind (html, options) {
  if (!options) options = {}

  this.Query = Query
  this.$node = $(html)
  this.directives = merge({}, defaultDirectives, options.directives)
  this.postDirectives = options.postDirectives || {}
  this.data = {}
  this.templates = {}
  this.logging = options.logging

  /**
   * Checks if a directive referenced by the directiveKey should be applied
   * @param directiveKey - the reference to the directive
   * @param value - the data value
   * @param key - the data key
   * @param context - context object
   * @returns {*}
   */
  this.match = function(directiveKey, value, key, context) {
    let directive = normalizeDirective(this.directives[directiveKey], directiveKey)
    let params = [value, key, context]
    let match = directive.match

    if (typeof match !== 'function') {
      match = function(value, key) {
        return key === directive.bind
      }
    }

    if (match.apply(this, params))
      return directive
    else
      return false
  }

  /**
   * NOTE: we purposely don't include binding by class be default, it's good to be explicit
   * - don't use class for styling and template binding
   * @param $context
   * @param key
   * @returns {*}
   */
  this.bindEach = function($template, onMatch) {
    if (!$template || !($template instanceof Query))
      return $template

    let bindKey = $template.attr('data-bind') || $template.attr('id')

    if (bindKey !== undefined) {
      this.log('- [BINDING]', bindKey)
      let result = onMatch($template, bindKey)

      $template = result.$rendered

      if (result.next === false)
        return $template
    }

    // Walk down the DOM tree and render any directives we see
    // ref: https://github.com/angular/angular.js/blob/master/src/ng/compile.js
    let $newChildren = $template.getChildren().map($child => this.bindEach($child, onMatch))
    $template.html($newChildren)

    // Note: on the 2nd iteration, it can return an array because we call $template.html instead of replaceWith
    return $template
  }

  this.memoized = function(cacheKey, $template) {
    let $memoized

    if (this.templates[cacheKey]) {
      $memoized = this.templates[cacheKey].clone()

    } else {
      $memoized = $template ? $template.clone() : $template
      this.templates[cacheKey] = $memoized
    }

    return $memoized
  }
}

Bind.prototype = {

  render: function(data) {
    this.data = data
    // clear the layout
    this.templates = {}

    let $node = this._render(this.$node, data)

    return $node
  },

  // Recursive render function
  _render: function($template, data, path) {
    this.log('[RENDERING]', path || '', data)

    if (!path)
      path = ''

    if (!data || (typeof data != 'object'))
      data = {}

    // TODO: this needs to support the case where there are multiple templates for the same bind key?
    // NOTE: if there are two different templates using the same binding key, it'll need to be aliased
    let self = this,
      $node = this.memoized(path || '__LAYOUT', $template)

    // First process all the meta directives
    for (let key in metaDirectives) {
      let directive = normalizeDirective(metaDirectives[key], key)
      let value = data[key]
      let context = { key, path, $node, data, value }
      if (value !== undefined)
        $node.replaceWith(directive.render.call(this, $node, value, context))
    }

    $node.replaceWith(this.bindEach($node, onMatch))

    // Process any post directives
    for (let key in this.postDirectives) {
      let directive = normalizeDirective(this.postDirectives[key], key)
      let value = data[key]
      let context = { key, path, $node, data, value }
      if (value !== undefined)
        $node.replaceWith(directive.render.call(this, $node, value, context))
    }

    this.log('[RENDERED] ', path || '')

    return $node

    // TODO move these to static functions
    // TODO need to clone $target?
    function onMatch($target, key) {
      let value = data[key]
      let result = { $rendered: $target }

      if (value !== undefined) {
        let context = { key, path, $node, data, value }
        let directive = matchDirective(key, context)
        let $rendered = directive.render.call(self, $target, context.value, context)
        result = { $rendered }

        self.log('- [BOUND]  ', directive.bind)

        // Recursive directives handle rendering of their own children, continue to next tree
        if (directive.recursive)
          result.next = false
      }

      return result
    }

    /**
     * Match which directives are loaded first
     * - custom directives defined by the user are always matched first
     * @param key
     * @param context
     * @returns {} - matched directive
     */
    function matchDirective(key, context) {
      let value = context.data[key]

      let defaultDirectiveKeys = Object.keys(self.directives)
        .filter(key => defaultDirectives[key])

      let userDirectiveKeys = Object.keys(self.directives)
        .filter(key => defaultDirectives[key] === undefined)

      // Ensure that the user directives are always matched before default directives
      let directiveKeys = userDirectiveKeys.concat(defaultDirectiveKeys)

      for (let directiveKey of directiveKeys) {
        let directive = self.match(directiveKey, value, key, context)
        if (directive && directive.render) return directive
      }
    }
  },

  log: function() {
    if (console && console.info && this.logging)
      console.info.apply(this, arguments)
  }
}

function compact(array) {
  if (!array || !(array instanceof Array)) return []

  return array.filter(el => el !== null && el !== undefined && el !== '')
}


/**
 * Normalize directive shorthand declarations to standard syntax
 * @param directive
 * @param directiveKey
 * @returns {} - normalized directive
 */
function normalizeDirective(directive, directiveKey) {
  if (typeof directive === 'function')
    directive = { render: directive }

  if (!directive.render)
    throw new Error('Invalid Directive: '+directive)

  if (isNaN(directive.priority))
    directive.priority = 1

  if (!directive.bind)
    directive.bind = directiveKey //directive.render.name

  if (!directive.bind)
    throw new Error('Invalid Directive '+directive.render)

  return directive
}

module.exports = Bind