'use strict';

const assert    = require('chai').assert;
const fs        = require('fs')
const Bind      = require('../lib')

const React = require('react')
const createReactComponents = require('html-to-react').createReactComponents(React)
const renderToStaticMarkup = require('react-dom/server').renderToStaticMarkup

const Parser = require('html-to-react').Parser
const parser = new Parser()

function assertEqual(componentTree, html) {
  let components = createReactComponents(componentTree)
  let output = renderToStaticMarkup(components)
  assert.equal(output, html)
}

describe('render', function (){

  it('should ignore null data and non matching data', function (){
    let html = '<title data-bind="title"></title>'
    let template = new Bind(parser.parse(html))

    let $rendered = template.render({
      foobar: 'baz',
      '@bad-key': ''
    })
    assertEqual($rendered, html);
  })

  it('should match and render data-bind attributes', function (){
    let html = '<title data-bind="title"></title>'
    let template = new Bind(parser.parse(html))

    let $rendered = template.render({title: 'hello'})
    assertEqual($rendered, '<title data-bind="title">hello</title>');
  })

  it('should match and render attributes and html', function (){
    let html = '<body><section id="content"></section></body>'
    let template = new Bind(parser.parse(html))

    let $rendered = template.render({content: {_html: 'foobar'}})
    assertEqual($rendered, '<body><section id="content">foobar</section></body>');

    // TODO support attributes
    //let $rendered = template.render({content: {_className: 'main', _id: 'alt1'}})
    //assertEqual($rendered, '<body><section id="alt1" class="main"></section></body>');

    // Should be able to re-render
    //$rendered = template.render({content: {_className: 'main row'}})
    //assertEqual($rendered, '<body><section id="content" class="main row"></section></body>');

    // Should be able to set null
    //render($html, {content: {_html: '<div>foobar</div>', _class: null}})
    //assert.equal($html.toString(), '<body><section id="content"><div>foobar</div></section></body>');
  })

  it('should match nested data', function() {
    let html = '<body>' +
      '<section id="content">' +
        '<article id="post"></article>' +
        '<article data-bind="post2"></article>' +
      '</section>' +
      '<div id="footer"></div>' +
    '</body>'
    let template = new Bind(parser.parse(html))

    let $output = template.render({
      content: {
        post: 'foo',
        post2: 'bar'
      },
      footer: 'footer'
    })

    // TODO bug: it shouldn't try to bind on content again on the recursive call? or just dont' support self finds, see logs
    assertEqual($output,
    '<body>' +
      '<section id="content">' +
        '<article id="post">foo</article>' +
        '<article data-bind="post2">bar</article>' +
      '</section>' +
      '<div id="footer">footer</div>' +
    '</body>')

    // Ensure that it can be re-rendered consistently
    $output = template.render({
      content: {post2: 'post2'}
    })
    assertEqual($output,
    '<body>' +
      '<section id="content">' +
        '<article id="post"></article>' +
        '<article data-bind="post2">post2</article>' +
      '</section>' +
      '<div id="footer"></div>' +
    '</body>')
  })

  it('should match and render an array of items', function (){
    let html = '<ul>' +
      '<li data-bind="list-item"><i data-bind="item"></i></li>' +
    '</ul>'

    let template = new Bind(parser.parse(html))
    let data = {
      'list-item': [
        {item: 'first'},
        {item: 'second'}
      ]
    }

    let $output = template.render(data)

    assertEqual($output, '<ul>' +
      '<li data-bind="list-item"><i data-bind="item">first</i></li>' +
      '<li data-bind="list-item"><i data-bind="item">second</i></li>' +
    '</ul>');

    // Render it again to test that re-rendering works
    $output = template.render(data)
    assertEqual($output, '<ul>' +
      '<li data-bind="list-item"><i data-bind="item">first</i></li>' +
      '<li data-bind="list-item"><i data-bind="item">second</i></li>' +
    '</ul>');
  })

  it('should use proper scoping', function() {
    let html = '<body>' +
      '<section id="content">' +
        '<article data-bind="post"></article>' +
      '</section>' +
      '<div data-bind="post"></div>' +
      '<div data-bind="post"></div>' +
    '</body>'
    let template = new Bind(parser.parse(html))

    // test that
    let $output = template.render({
      content: {},
      post: 'bar'
    })
    assertEqual($output, '<body>' +
      '<section id="content">' +
        '<article data-bind="post"></article>' +
      '</section>' +
      '<div data-bind="post">bar</div>' +
      '<div data-bind="post">bar</div>' +
    '</body>')
  })

  it('should support meta directives and template rendering', function() {
    let html = '<body>' +
      '<section id="section"></section>' +
      '<template id="section-tmplt">' +
        '<article id="article"></article>' +
      '</template>'
    '</body>'

    let template = new Bind(parser.parse(html))
    let $output = template.render({
      section: {
        _template: 'section-tmplt',
        article: 'foobar'
      }
    })

    // TODO cleanup: should remove the templates from output
    assertEqual($output, '<body>' +
      '<section id="section">' +
        '<article id="article">foobar</article>' +
      '</section>' +
      '<template id="section-tmplt">' +
        '<article id="article"></article>' +
      '</template>' +
    '</body>')
  })

  it('should support custom directives', function () {
    let html = '<ul data-bind="_yield"></ul>'
    let data = {
      _yield: [
        'first',
        'second'
      ]
    }
    let directives = {
      _yield: function ($target, value, context) {
        return context.$node.text(value.join(', '))
      }
    }
    let template = new Bind(parser.parse(html), directives)

    let $output = template.render(data)
    assertEqual($output, '<ul data-bind="_yield">first, second</ul>');
  })


  // TODO write a more extensive test, test all of the features working together
  it('should support array of custom directives', function () {
    let html = '<body>' +
      '<section data-bind="sections"><i data-bind="element"></i></section>' +
    '</body>'

    let data = {
      sections: [
        { element: 50},
        { element: 100}
      ]
    }

    let directives = {
      element: ($target, value) => {
        return $target.css({width: `${value}%`})
      }
    }

    let template = new Bind(parser.parse(html), directives)

    let $output = template.render(data)
    assertEqual($output, '<body>' +
      '<section data-bind="sections">' +
        '<i data-bind="element" style="width:50%;"></i>' +
      '</section>' +
      '<section data-bind="sections">' +
        '<i data-bind="element" style="width:100%;"></i>' +
      '</section>' +
    '</body>');
  })

  xit('should follow proper render and matching order', function (){
    let html = '<body id="content"><section id="section"></section></body>'
    let template = new Bind(parser.parse(html), [
      {
        priority: 100,
        match: function(value, key) {
          return key === '_clear'
        },
        render: function _clear($target, value, context) {
          return context.$node.html(value)
        }
      }
    ])

    let $output = template.render({
      section: {_class: 'main', _id: 'alt1'},
      _clear: ''
    })
    assert.equal($output.toString(), '<body id="content"></body>');
  })

  xit('should deliver the correct path to the lifecycle hook', function() {
    let html = '<body>' +
      '<section data-bind="positions">' +
        '<header id="header"><i id="icon"></i></header>' +
        '<article data-bind="contents">' +
          '<header data-bind="header"></header>' +
        '</article>' +
      '</section>' +
      '</body>'
    let fullPaths = []
    let template = new Bind(parser.parse(html), null, function($target, bind, context) {
      fullPaths.push(_.compact([context.path, context.key]).join('.'))
    })

    let $output = template.render({
      positions: {
        header: {icon: 'star'},
        contents: [
          {header: 'one'},
          {header: 'two'}
        ]
      }
    })
    assert.equal($output.toString(), '<body>' +
      '<section data-bind="positions">' +
        '<header id="header"><i id="icon">star</i></header>' +
        '<article data-bind="contents">' +
          '<header data-bind="header">one</header>' +
        '</article>' +
        '<article data-bind="contents">' +
          '<header data-bind="header">two</header>' +
        '</article>' +
      '</section>' +
    '</body>')

    assert.deepEqual(fullPaths, [
      'positions.header.icon',
      'positions.header',
      'positions.contents.0.header',
      'positions.contents.1.header',
      'positions.contents',
      'positions'
    ])
  })

  xit('should render a simple layout with partials', function (){
    const layout = fs.readFileSync(__dirname+'/fixtures/rendering/_layout.html', 'utf-8')
    const page = fs.readFileSync(__dirname+'/fixtures/rendering/index.html', 'utf-8')
    const result = fs.readFileSync(__dirname+'/fixtures/rendering/results/index.html', 'utf-8')

    let pageData = require('./fixtures/rendering/_index.json')

    // sideload the layout and partials
    pageData._layout = layout
    pageData._yield._template = page

    // render
    let html = '<html></html>'
    let template = new Bind(parser.parse(html))
    template.render(pageData)

    let compare = hiff.compare($template.toString(), result)
    assert.isFalse(compare.different);
  })

})